#include<stdio.h>
#include<math.h>
#include<stdlib.h>
#include<string.h>
#include<time.h>
#include<curand.h>
#include<curand_kernel.h>
#include<limits.h>

#ifndef robonum
#define robonum 4//i/no. of robot
#endif
#ifndef resnum
#define resnum 3 //no. of resources
#endif
#ifndef tasknum
#define tasknum 3 //no. of tasks
#endif
#ifndef OBJNUM
#define OBJNUM 2 // no. of objectives
#endif
#ifndef PM
#define PM 0.1 //per gene mutation probibility
#endif
#ifndef CHROMSIZE
#define CHROMSIZE (tasknum*robonum)//tasknum*robonum//size of the chromosome
#endif
#ifndef MINMAX
#define MINMAX 0 //0=min , !0 =max
#endif
#define LARGE 20000000000
#ifndef MAX_POP
#define MAX_POP 4 //no. of populations / threads
#endif
#define MAX_POP2 MAX_POP*2 //population of  union pop
#ifndef ITERATIONS
#define ITERATIONS 100000 // no. of interations 
#endif
#ifndef MAX_SOL
#define MAX_SOL 1+tasknum //maximum no. of solutions
#endif
#ifndef CONVERGENCE
#define CONVERGENCE 500//post this value the front shall treated to be converged
#endif
FILE *file;//file for displaying the results
FILE *FRM,*FTM,*FTDM;
int RM[robonum*resnum],TM[tasknum*resnum],TDM[tasknum*robonum];
typedef struct solution
{
	int chrom[tasknum*robonum];
	float obj[OBJNUM];

}individual;

//structure representing population (array of individual)

typedef struct archive_array
{

	individual ind[MAX_POP];
}population;

//strucure representing union population

typedef struct archive_array2

{	individual ind[2*MAX_POP];
}population2;

//Front front convergence testing
individual old_front1[MAX_SOL];
individual current_front1[MAX_SOL];
int old_len_front1=0;
int count_converge=0;//for no. of repetetive fronts
__host__ void input(FILE *f,int size,int *array)
{
	int n=0,i=0;
	while(fscanf(f,"%d,",&n)>0)
	{
		array[i++]=n;
	}
}


__device__ void init(individual *c,curandState *state)

{
	int r_num;
	int idx=threadIdx.x;

	

	for(int i=0;i<tasknum*robonum;i++)
		{
			float randn=curand_uniform(&state[idx]);
	
			if(randn<=(tasknum/((float)(robonum))))

				r_num=1;
			else
				r_num=0;

			c->chrom[i]=r_num;
		}

}





//evaluates the generated the chromosme against each of the objectives and generates the fitness value corresponding to each objective
//objective 1	: no. of incomplete tasks
//objective 2  	: cumulative distance travelled by all the robots involved
//negative values of each are taken for maximization

__device__ void evaluate(individual *chrm,int *robomat, int *taskmat,int *taskdistmat) 

{

	int taskstat[tasknum];//status of the tasks : complete/incomplete
	int sum,temp=0;
	chrm->obj[0]=0;//no. of incomplete tasks, initially it's assumed that all the tasks are complete
	chrm->obj[1]=0;//total distance travelled by each of the robots for each tasks

	//evaluation of objective 1, task completeion status
	
	for(int i=0;i<tasknum;i++){
		temp=0;
		for(int m=0;m<resnum;m++){
			sum=0;
			for(int j=0;j<robonum;j++){
				if(chrm->chrom[i*robonum+j]==1)

					sum+=robomat[j*resnum+m];

			}
		
		if(sum<taskmat[i*resnum+m])
			{
				   temp=1; break;
				
			}
		}
		if(temp==0)
			taskstat[i]=1;
		else
		{
			chrm->obj[0]++;	
			taskstat[i]=0;
		}

	}
	chrm->obj[0]*=-1;
	
	//evaluation of objective 2, distance
	
	for(int i=0;i<tasknum;i++){

		if(taskstat[i]==1)
			for(int j=0;j<robonum;j++)
				if(chrm->chrom[i*robonum+j]==1)
				{
					chrm->obj[1]+=taskdistmat[i*robonum+j];
			
				}

	}
	chrm->obj[1]*=-1;
	

}




__device__ void check(individual *chrm, curandState *state)//Checks if multiple tasks are allocated to a single robot, resolves the ambiguity by randoming alloting one of the [previously allocated tasks. Also randomly allocates a task to a robot ewich hasn't been allocated any task previously.
{	
	int temp[tasknum];//temporary array stores the location of the  tasks allocated to a robot
	int count=0; // temporary storage for number of tasks for a robot
	int idx=threadIdx.x;


	for(int i=0;i<robonum;i++){
		count=0;
		for(int j=0;j<tasknum;j++){

			if(chrm->chrom[j*robonum+i]==1){
				chrm->chrom[j*robonum+i]=0;
				count++;
				temp[count-1]=j*robonum+i;
			}}
		if(count!=0)
			chrm->chrom[temp[((int)(count*curand_uniform(&state[idx])))]]=1;
		else
			chrm->chrom[(((int)(tasknum*curand_uniform(&state[idx]))))*robonum+i]=1;
	}


}

__device__ __host__ int compare_max(individual *s1,individual *s2)

{	
if((s1->obj[0]>s2->obj[0] && s1->obj[1]<s2->obj[1]) || (s1->obj[0]<s2->obj[0]&&s1->obj[1]>s2->obj[1]))
	return 0;
else if(s1->obj[0]>=s2->obj[0]&&s1->obj[1]>=s2->obj[1])
	return 1;
else
	return -1;
}


__device__ void mutate(individual *c,int *robomat, int *taskmat,int *taskdistmat,curandState *state)
{
	int robot_mutate=0;
	int mutate_ind;
	int idx=threadIdx.x;
	individual temp;
	memcpy(&temp,c,sizeof(individual));
	for(int i=0;i<ceil(PM*CHROMSIZE);i++)
	{
		mutate_ind=((int)(CHROMSIZE*curand_uniform(&state[idx])));
		robot_mutate=mutate_ind%robonum;
		if(temp.chrom[mutate_ind]==0)
		{
			for(int j=0;j<tasknum;j++)
				temp.chrom[j*robonum+robot_mutate]=0;
		temp.chrom[mutate_ind]=1;
		}
		else
		{
			temp.chrom[mutate_ind]=0;
			temp.chrom[((int)(tasknum*(curand_uniform(&state[idx]))))*robonum+robot_mutate]=1;
		}
	}

	evaluate(&temp,robomat,taskmat,taskdistmat);
	if(compare_max(c,&temp)!=1)
	memcpy(c,&temp,sizeof(individual));

	

}

// A utility function to swap two elements
void swap ( int* a, int* b )
{
    int t = *a;
    *a = *b;
    *b = t;
}
 
// This function is same in both iterative and recursive
int partition (float *arr,int *indices, int l, int h)
{
    	float x=arr[h];
    int i = (l - 1);
 	float temp;
    for (int j = l; j <= h- 1; j++)
    {
        if (arr[j] <= x)
        {
            i++;
		temp=arr[i];
		arr[i]=arr[j];
		arr[j]=temp;
            swap (&indices[i], &indices[j]);
        }
    }
	temp=arr[i+1];
	arr[i+1]=arr[h];
	arr[h]=temp;
    swap (&indices[i + 1], &indices[h]);
    return (i + 1);
}
 
// A[] --> Array to be sorted, l  --> Starting index, h  --> Ending index 
void quick_sort2(float *arr,int *indices, int l, int h)
{
    // Create an auxiliary stack
    int stack[ h - l + 1 ];
  
    // initialize top of stack
    int top = -1;
 
    // push initial values of l and h to stack
    stack[ ++top ] = l;
    stack[ ++top ] = h;
 
    // Keep popping from stack while is not empty
    while ( top >= 0 )
    {
        // Pop h and l
        h = stack[ top-- ];
        l = stack[ top-- ];
 
        // Set pivot element at its correct position in sorted array
        int p = partition( arr,indices, l, h );
 
        // If there are elements on left side of pivot, then push left
        // side to stack
        if ( p-1 > l )
        {
            stack[ ++top ] = l;
            stack[ ++top ] = p - 1;
        }
 
        // If there are elements on right side of pivot, then push right
        // side to stack
        if ( p+1 < h )
        {
            stack[ ++top ] = p + 1;
            stack[ ++top ] = h;
        }
    }
}
int partition1(population2 *arr,int *indices, int l, int h,int obj_n)
{

	individual x=arr->ind[h];
    int i = (l - 1);
 	individual temp;
    for (int j = l; j <= h- 1; j++)
    {
        if (arr->ind[j].obj[obj_n] <= x.obj[obj_n])
        {
            i++;
		temp=arr->ind[i];
		arr->ind[i]=arr->ind[j];
		arr->ind[j]=temp;
            swap (&indices[i], &indices[j]);
        }
    }
	temp=arr->ind[i+1];
	arr->ind[i+1]=arr->ind[h];
	arr->ind[h]=temp;
    swap (&indices[i + 1], &indices[h]);
    return (i + 1);
}
 
// A[] --> Array to be sorted, l  --> Starting index, h  --> Ending index 
void quick_sort(population2 arr,int *indices, int l, int h, int obj_n=0)
{
    // Create an auxiliary stack
    int stack[ h - l + 1 ];
  
    // initialize top of stack
    int top = -1;
 
    // push initial values of l and h to stack
    stack[ ++top ] = l;
    stack[ ++top ] = h;
 
    // Keep popping from stack while is not empty
    while ( top >= 0 )
    {
        // Pop h and l
        h = stack[ top-- ];
        l = stack[ top-- ];
 
        // Set pivot element at its correct position in sorted array
        int p = partition1(&arr,indices, l, h,obj_n );
 
        // If there are elements on left side of pivot, then push left
        // side to stack
        if ( p-1 > l )
        {
            stack[ ++top ] = l;
            stack[ ++top ] = p - 1;
        }
 
        // If there are elements on right side of pivot, then push right
        // side to stack
        if ( p+1 < h )
        {
            stack[ ++top ] = p + 1;
            stack[ ++top ] = h;
        }
    }
}


int partition3(population2 *arr,int *d,int *indices, int l, int h,int obj_n)
{

 
	individual x=arr->ind[h];
    int i = (l - 1);
 	individual temp;
    for (int j = l; j <= h- 1; j++)
    {
        if (arr->ind[j].obj[obj_n] <= x.obj[obj_n])
        {
            i++;
		temp=arr->ind[i];
		arr->ind[i]=arr->ind[j];
		arr->ind[j]=temp;
            swap (&indices[i], &indices[j]);
		swap(&d[i],&d[j]);
        }
    }
	temp=arr->ind[i+1];
	arr->ind[i+1]=arr->ind[h];
	arr->ind[h]=temp;
    swap (&indices[i + 1], &indices[h]);
	swap(&d[i+1],&d[h]);
    return (i + 1);
}
 
// A[] --> Array to be sorted, l  --> Starting index, h  --> Ending index 
void quick_sort3(population2 arr,int *d,int *indices, int l, int h, int obj_n)
{
    // Create an auxiliary stack
    int stack[ h - l + 1 ];
  
    // initialize top of stack
    int top = -1;
 
    // push initial values of l and h to stack
    stack[ ++top ] = l;
    stack[ ++top ] = h;
 
    // Keep popping from stack while is not empty
    while ( top >= 0 )
    {
        // Pop h and l
        h = stack[ top-- ];
        l = stack[ top-- ];
 
        // Set pivot element at its correct position in sorted array
        int p = partition3(&arr,d,indices, l, h,obj_n );
 
        // If there are elements on left side of pivot, then push left
        // side to stack
        if ( p-1 > l )
        {
            stack[ ++top ] = l;
            stack[ ++top ] = p - 1;
        }
 
        // If there are elements on right side of pivot, then push right
        // side to stack
        if ( p+1 < h )
        {
            stack[ ++top ] = p + 1;
            stack[ ++top ] = h;
        }
    }
}


__host__ void select(population *old_pop,population *new_pop)
{//selects and generates a new population using tournament selection



	int pop=MAX_POP;
	int flag[pop],ind1,ind2;
	int selected_ind=0,temp,temp_eq;
	int r1,r2; 
	bool break1=false,break2=false;
	for(int i=0;i<pop;i++)
		flag[i]=0;
	int counter=0;//counts the no. of matches undertaken
	int val=2; //for match in phase 1 & phase 2
	for(int i=0;i<pop;i++)
	{	break1=break2=false;
		temp=0;
		temp_eq=0;
		ind1=0;
		ind2=0;
		r1=rand()%(pop-i);
		r2=(pop-i-1)!=0?rand()%(pop-i-1):0;
		
		while(true)
		{
			if(counter==pop/2)
				val=10;//for phase 2 of matches when for all flag>=2(10 is a random value here)

			
			if(!break1)
			{

				if(flag[r1]<3 && flag[r1]!=val)
				{
					ind1=r1;
					flag[ind1]+=3;
					break1=true;
				}
				else if(r1+1<pop)
					r1++;
				else r1=0;
				}
			if(!break2)
		{
				if(flag[r2]<3 && flag[r2]!=val)
				{
					ind2=r2;
					flag[ind2]+=3;
					break2=true;
				}
				else if(r2+1<pop)
					r2++;
				else r2=0;
			}
			if(break1 && break2)
			{	counter++;	
				flag[ind1]-=1;
				flag[ind2]-=1;
				break;
			}
		}

		for(int j=0;j<OBJNUM;j++)
		{
			if(old_pop->ind[ind1].obj[i]>old_pop->ind[ind2].obj[i])
				temp++;
			else if(old_pop->ind[ind1].obj[i]==old_pop->ind[ind2].obj[i])
				temp_eq++;
		}
		if(compare_max(&(old_pop->ind[ind1]),&(old_pop->ind[ind2]))==0)
		{	//randomly seelcts either of ind1 and ind2
			memcpy(&(new_pop->ind[selected_ind]),&(old_pop->ind[(rand()%2==0)?ind1:ind2]),sizeof(individual));
			selected_ind++;
			//printf(" max min \n %d",selected_ind-1);
		}
		else 
		{	
			memcpy(&(new_pop->ind[selected_ind]),&(old_pop->ind[(compare_max(&(old_pop->ind[ind1]),&(old_pop->ind[ind2])))==1?ind1:ind2]),sizeof(individual));
			selected_ind++;
		}
	}

}

int* front(int *indices,population2 fit,int *length)
{
	

	int i=0,j=0;
	int temp=0;
	int *TOP=NULL,*BOT=NULL,len_T=0,len_B=0,flag=0;
	int *TOPind;
	TOPind=(int *)malloc(MAX_POP2*sizeof(int));
	int indices_t[MAX_POP2],indices_b[MAX_POP2];
	int len=*length;
	if(len==1)
	{	 
		return indices;
	}

	for(i=0;i<MAX_POP2;i++)
        {
       		indices_t[i]=-1;
			indices_b[i]=-1;
	}

	int mid=(int)floorl(len/2);
	len_T=mid;
	len_B=len-mid;

	for(i=0;i<len_T;i++)
	{
		indices_t[i]=indices[i];
	}
                                               
	for(i=0;i<len_B;i++)
		indices_b[i]=indices[len_T+i];
	                   
	TOP=front(indices_t,fit,&len_T);
	for(i=0;i<MAX_POP2;i++)
	TOPind[i]=TOP[i];
	BOT=front(indices_b,fit,&len_B);
	TOP=TOPind;
	for(i=0;i<len_B;i++)
	{
		flag=1;
		for(j=0;j<len_T;j++)
		{
			if(fit.ind[BOT[i]].obj[1]<=fit.ind[TOP[j]].obj[1])
			{	
				flag=0;
				break;
			}
			else if((fit.ind[BOT[i]].obj[0]==fit.ind[TOP[j]].obj[0]) && (fit.ind[BOT[i]].obj[1]>fit.ind[TOP[j]].obj[1]))
			{	    

       				flag=0;
 	      			temp=BOT[i];
       				TOP[j]=temp;
       		  					
      			}
      		}
      
      		if(flag==1 && fit.ind[BOT[i]].obj[1]!=fit.ind[TOP[j]].obj[1])
      		{	
      			TOP[len_T++]=BOT[i];
      		}
   	}
       		indices=TOP;
		*length=len_T;
      	
	 	return indices;

}


void min_max(population2 pop2, int *indices,int len,int obj_n,float *min, float *max)
{
	*min=pop2.ind[indices[0]].obj[obj_n];
	*max=pop2.ind[indices[0]].obj[obj_n];
	for(int i=1;i<len;i++)
	{
		if(*max<pop2.ind[indices[i]].obj[obj_n])
			*max=pop2.ind[indices[i]].obj[obj_n];
		if(*min>pop2.ind[indices[i]].obj[obj_n])
			*min=pop2.ind[indices[i]].obj[obj_n];
	
	}


}



int *crowd_dist(population2 pop2,int *front,int len)
{	

	int temp;
	int *indices=NULL;
	float max1,min1,max2,min2; //for  storing the maximum and minimum values corresponding to objective 1 & objective 2
	indices=(int *)malloc(len*sizeof(int));
	for(int i=0;i<len;i++)
		indices[i]=i;

	float d[len];
	int *sorted_indices;
	sorted_indices=(int *)malloc(len*sizeof(int));
	for(int i=0;i<len;i++)
		sorted_indices[i]=i;
	min_max(pop2,front,len,0,&min1,&max1);// call for computing the max /min vals of corresponding to objective in 1 among the indices in the array "front"
	min_max(pop2,front,len,1,&min2,&max2);// for objective 2
	float max_obj[OBJNUM]={max1,max2};     	
	float min_obj[OBJNUM]={min1,min2};	
	for(int i=0;i<len;i++)
		d[i]=0;
	for(int j=0;j<OBJNUM;j++)
	{
		quick_sort3(pop2,front,indices,0,len-1,j);
		d[indices[0]]=INT_MAX;
		d[indices[len-1]]=INT_MAX;

		for(int i=1;i<=len-2;i++)
		{ 	if(max_obj[j]==min_obj[j])
				d[indices[i]]=d[indices[i]]+2000;
			else
				d[indices[i]]=d[indices[i]]+(pop2.ind[indices[i+1]].obj[j]-pop2.ind[indices[i-1]].obj[j])/(max_obj[j]-min_obj[j]);

		}

	}
	quick_sort2(d,sorted_indices,0,len-1);
	for(int i=0;i<len/2;i++)                              	
	{                                                             	
		temp=sorted_indices[i];
		sorted_indices[i]=sorted_indices[len-1-i];
		sorted_indices[len-1-i]=temp;
		
	
	}


	free(indices);
	return sorted_indices;

}

__host__ int compare_front(individual *old_front,individual *current_front,int old_len, int current_len)
{
int matches=0;//matches for objectives
int row_match=0;// matching of an  entire solutions
if(old_len==current_len)
{
	for(int i=0;i<old_len;i++)
	{
		for(int j=0;j<current_len;j++)
		{	matches=0;
			for(int k=0;k<OBJNUM;k++)
				if(old_front[i].obj[k]==current_front[j].obj[k])
					matches++;
				else break;
			if(matches==OBJNUM)
			{	row_match++;
				break;	
			}
		}
	if(i+1!=row_match)
		break;
	}	
	if(row_match==old_len)
		return 1;
	else return 0;
}

else
	return 0;


}



__host__ population *sort(population *old_pop, population *mute_pop)

{/* the function combines the old and the mutated population and performs a non-dominated sorting on them, it essentially starts with sorting the population indices with respected to its first objective using quicksort followed by generation of front and computation of crowding distances using corresponding function wherever necessar*/	


	int *indices=NULL,*indices_covered=NULL,temp,counter=0;//indices- > points to an array of indices of newly formed union population (mutated pop+ old pop)
								// indices_covered -> contained indices already been covered by any of the fronts
        indices=(int *)malloc(2*MAX_POP*sizeof(int));
	
	int *len=NULL;// length of the array to be sorted (to be used while applying top down approach for sorting)

	len=(int *)malloc(sizeof(int));
	*len=MAX_POP*2;
	for(int i=0;i<2*MAX_POP;i++)
		indices[i]=i;//indices initialised with the 0,1,2,3.... i.e. the inital  state
	int popsize=MAX_POP;
	int new_popsize=0;//current  size of the sorted population ( shall change as when new fronts are added) 
	int front_size=0; //size of the current front

	indices_covered=(int *)malloc(2*MAX_POP*sizeof(int));// indices added to the new population (p(t1+1))
	int *indices_ordered=NULL; 
	indices_ordered=(int *)malloc(2*MAX_POP*sizeof(int));// an ordered list of indices {1,2,3,4...}
	int *indices_crowd=NULL;//list of the indices for which crowding distance is to be calculated
	population *next_pop;	//shall point to the new sorted popul
	next_pop=(population *)malloc(sizeof(population));
	population2 union_pop;

	memcpy(&(union_pop.ind[0]),&(old_pop->ind[0]),MAX_POP*sizeof(individual));
	memcpy(&(union_pop.ind[MAX_POP]),&(mute_pop->ind[0]),MAX_POP*sizeof(individual));

	quick_sort(union_pop,indices,0,2*MAX_POP-1);
	for(int i=0;i<(2*MAX_POP)/2;i++)
	{
		temp=indices[i];
		indices[i]=indices[2*MAX_POP-1-i];
		indices[2*MAX_POP-1-i]=temp;
		indices_ordered[i]=indices[i];
		indices_ordered[2*MAX_POP-1-i]=indices[2*MAX_POP-1-i];
		
	}

	int flag=0;
	while(new_popsize<popsize)
	{
		     	counter=0;


		indices=front(indices,union_pop,len);
		flag++;//for printing front 1 only
		front_size=*len;
		new_popsize+=front_size;
		if(flag==1)
		{	
		/*	
			fprintf(file,"\n%s %d \n","Objective values of front 1, front length = ",*len);
			fprintf(file,"%s\t%s\t%s\n","Objective 1","Objective 2","Index");
			

*/
			for(int h=0;h<*len;h++)
			{
			//	fprintf(file,"%f\t%f\t%d\n",union_pop.ind[indices[h]].obj[0],union_pop.ind[indices[h]].obj[1],indices[h]);			
				memcpy(&current_front1[h],&union_pop.ind[indices[h]],sizeof(individual));
			}
			if(compare_front(old_front1,current_front1,old_len_front1,*len)==1)
				count_converge++;
			else
				count_converge=0;
			old_len_front1=*len;
			memcpy(old_front1,current_front1,sizeof(current_front1));
		}
		if(new_popsize<=popsize)
		{	for(int i=new_popsize-front_size;i<new_popsize;i++)
				indices_covered[i]=indices[i-(new_popsize-front_size)];
		}		
		else
			break;		

		for(int i=0;i<2*MAX_POP;i++)
		{	temp=0;
			for(int j=0;j<new_popsize;j++)
				if(indices_ordered[i]==indices_covered[j])
					temp=1;
			if(temp!=1)
			{
				indices[counter]=indices_ordered[i];
				counter++;
			}
		}
		*len=counter;

	}//mechanism for crowding distances
	if(new_popsize>popsize)
	{
		indices_crowd=crowd_dist(union_pop,indices,front_size);
		for(int i=0;i<popsize-(new_popsize-front_size);i++)//for adding the deficient indices
			indices_covered[new_popsize-front_size+i]=indices_crowd[i];
	}
	for(int i=0;i<MAX_POP;i++)
		next_pop->ind[i]=union_pop.ind[indices_covered[i]];
	free(len);
	free(indices);
	free(indices_ordered);
	free(indices_covered);
	if(new_popsize>popsize)
		free(indices_crowd);
		
	return next_pop;

}



__global__ void initialize(int *robomat,int *taskmat,int *taskdistmat,population *pop_in,curandState *state)//paralllely generates chromosomes and evaluate them against the objectives 
{


	
	int idx=threadIdx.x;//thread no.
	curand_init(idx,idx,idx,&state[idx]);
	init(&(pop_in->ind[idx]),state);//intializes the chromosomes with random values

	check(&(pop_in->ind[idx]),state);//check the generated chromosomes against the constraints
	evaluate(&(pop_in->ind[idx]),robomat,taskmat,taskdistmat);//evaluates the objective functions for each of the generated individuals
	
}

__global__ void mutate_pop(int *robomat,int *taskmat,int *taskdistmat,population *pop_mute,curandState *state)
{ 
	int idx=threadIdx.x;
	mutate(&(pop_mute->ind[idx]),robomat,taskmat,taskdistmat,state);

}




int main()
{	clock_t start=clock();
	population old_pop;//initalial population
	population new_pop;//population post tournament selection
	population mute_pop;// mutated version of new_pop
	int *d_in1;
	int *d_in2;
	int *d_in3;
	population *d_old_pop;	
	population *d_mute_pop;
	population *next_pop;
	int ARRAY_BYTES1=robonum*resnum*sizeof(int);
	int ARRAY_BYTES2=tasknum*resnum*sizeof(int);
	int ARRAY_BYTES3=robonum*tasknum*sizeof(int);
	curandState *d_state;
	char file_w_name[100];
	char file_res_robo[100];
	char file_res_task[100];
	char file_task_dist[100];

	cudaMalloc((void**)&d_in1,ARRAY_BYTES1);	
	cudaMalloc((void**)&d_in2,ARRAY_BYTES2);		
	cudaMalloc((void**)&d_in3,ARRAY_BYTES3);
	
	sprintf(file_res_robo,"Resource_Robot_%d_%d_1.txt",robonum,tasknum);
	sprintf(file_res_task,"Resource_Task_%d_%d_1.txt",robonum,tasknum);
	sprintf(file_task_dist,"Task_Robot_Distance_%d_%d_1.txt",robonum,tasknum);
	sprintf(file_w_name,"nsga_parallel_%d_%d_%d_%d_%d_%f_%d.txt",robonum,tasknum,resnum,ITERATIONS,MAX_POP,PM,CONVERGENCE);	
	FRM=fopen(file_res_robo,"r");
	input(FRM,robonum*resnum,RM);
	fclose(FRM);

	FTM=fopen(file_res_task,"r");
	input(FTM,tasknum*resnum,TM);
	fclose(FTM);

	FTDM=fopen(file_task_dist,"r");
	input(FTDM,robonum*tasknum,TDM);
	fclose(FTDM);
	file=fopen(file_w_name,"w");

	cudaMemcpy(d_in1,RM,ARRAY_BYTES1,cudaMemcpyHostToDevice);
	cudaMemcpy(d_in2,TM,ARRAY_BYTES2,cudaMemcpyHostToDevice);
	cudaMemcpy(d_in3,TDM,ARRAY_BYTES3,cudaMemcpyHostToDevice);

	cudaMalloc((void**)&d_old_pop,sizeof(population));
	cudaMalloc((void**)&d_state,MAX_POP*sizeof(curandState));
	cudaMalloc((void**)&d_mute_pop,sizeof(population));

	initialize<<<1,MAX_POP>>>(d_in1,d_in2,d_in3,d_old_pop,d_state);	// kernel call -> initalizes the old_pop population and evaluates the objective values after performing the check function
	cudaMemcpy(&old_pop,d_old_pop,sizeof(population),cudaMemcpyDeviceToHost);//copying of the old_pop from device to host memory
	fprintf(file,"Initial population fitness values \n\n");
	for(int i=0;i<MAX_POP;i++)
	{
		fprintf(file,"%f\t%f\n",old_pop.ind[i].obj[0],old_pop.ind[i].obj[1]);
	}



	int iter=0;
//iteration begin 		
	while(iter<ITERATIONS && count_converge!=CONVERGENCE)
	{
		iter++;
//		fprintf(file,"\n%s : %d\n","Iteration no.",iter);
//		printf("iteration no. %d \n",iter);
		select(&old_pop,&new_pop);//tournament selection , selected population is pointed to by the variable new_pop
    
  	        cudaMemcpy(d_mute_pop,&new_pop,sizeof(population),cudaMemcpyHostToDevice);
		mutate_pop<<<dim3(1,1,1),dim3(MAX_POP,1,1)>>>(d_in1,d_in2,d_in3,d_mute_pop,d_state);// kernel call for mutation of the selected population, along with its evaluation
		cudaMemcpy(&mute_pop,d_mute_pop,sizeof(population),cudaMemcpyDeviceToHost);
		next_pop=sort(&old_pop,&mute_pop);// performing of non-dominated sorting on the union of old and mutated population, the function return the sorted population for next iteration
		memcpy(&old_pop,next_pop,sizeof(old_pop)); // for next iteration


//		printf("\n ************************ iteration %d done*********************** \n",iter);
	}
	
	clock_t end=clock();
	float sec=(float)(end-start)/CLOCKS_PER_SEC;
//	printf("\nTime elapsed %f\n", sec);


//	printf("\n Convereged post %d iterations for a convergence of %d \n",iter,CONVERGENCE);



	
	fprintf(file,"\n%s %d \n","Front Length = ",old_len_front1);
	fprintf(file,"\n\n");
	
                                                                                                                                                 
                                                                                                                                                 
	for(int h=0;h<old_len_front1;h++)
	{
		fprintf(file,"%f\t%f\n",current_front1[h].obj[0],current_front1[h].obj[1]);			
		
	}
	fprintf(file,"\n\nIterations undertaken = %d",iter);
	fprintf(file,"\nConvergence Criteria = %d",CONVERGENCE);
	fprintf(file,"\nTime taken = %f",sec);




	fclose(file);

	cudaFree(d_in1);
	cudaFree(d_in2);
	cudaFree(d_in3);
	cudaFree(d_state);
	cudaFree(d_mute_pop);
	cudaFree(d_old_pop);



	return 0;

}


