#include<stdio.h>
#include<math.h>
#include<stdlib.h>
#include<curand.h>
#include<time.h>
#include<curand_kernel.h>
#ifndef robonum
#define robonum 20
#endif
#ifndef THRNUM
#define THRNUM 1
#endif
#ifndef resnum
#define resnum 4
#endif
#ifndef tasknum
#define tasknum 10
#endif
#ifndef OBJNUM
#define OBJNUM 2
#endif
#ifndef MAX_ARC
#define MAX_ARC (1+tasknum)		//maximum no. of elements in the arc1hive
#endif
#ifndef ITERATIONS
#define ITERATIONS 10000	//no. of iterations each thread shall conduct
#endif
#ifndef PM
#define PM 0.1			//per gene mutation probibility
#endif
#ifndef CHROMSIZE
#define CHROMSIZE (tasknum*robonum)	//tasknum*robonum//size of the chromosome
#endif
#ifndef MINMAX
#define MINMAX 0		//0=min , !0 =max
#endif
#ifndef DEPTH
#define DEPTH 5
#endif
//grid depth
#define LARGE 200000000000
#define MAX_LOC 3276
#ifndef MAX_POP
#define MAX_POP 4
#endif
#ifndef CONVERGENCE
#define CONVERGENCE 900
#endif
FILE *file;			//file for displaying results
FILE *FRM, *FTM, *FTDM;
int RM[robonum * resnum], TM[tasknum * resnum], TDM[tasknum * robonum];


typedef struct solution {
    int chrom[tasknum * robonum];
    double obj[OBJNUM];
    int grid_loc;
} sol;

typedef struct archive_array {
    sol arc[MAX_ARC];
} arch;

void input(FILE * f, int size, int *array)
{
    int n = 0, i = 0;
    while (fscanf(f, "%d,", &n) > 0) {
	array[i++] = n;
    }
}

__device__ int arc1length = 0;
__device__ sol arc1[MAX_ARC];
__device__ int change[THRNUM];	//substitute for static variable "change" in update_grid
__device__ int count=0;
__device__ int iterdone[THRNUM];


__device__ void init(sol * c, curandState * state)
{
    int r_num;
    int idx = threadIdx.x;
    c->grid_loc = 0;		//initalization of grid location with zero
    for (int i = 0; i < OBJNUM; i++)
	c->obj[i] = 0;

    for (int i = 0; i < tasknum; i++)
	for (int j = 0; j < robonum; j++) {
	    float randn = curand_uniform(&state[idx]);
	    if (randn <= (tasknum / (float) robonum))

		r_num = 1;
	    else
		r_num = 0;

	    c->chrom[i * robonum + j] = r_num;
	}

}


__device__ void print_eval(sol * s)
{
    for (int i = 0; i < OBJNUM; i++)
	printf("%g ", s->obj[i]);
    printf("n");
}

__device__ void add_to_arc1hive(sol * s)
{
    count = 0;	//if add_to_archive is being called for a particular thread
    arc1[arc1length] = *s;
    arc1length = arc1length + 1;
}



__device__ void evaluate(sol * chrm, int *robomat, int *taskmat, int *taskdistmat)	//evluates the genrated the chromosme against each of the objectives and generates the fitness value corresponding to each objective
{

    int sum, temp = 0;
    int taskstat[tasknum];

    chrm->obj[0] = 0;		//no. of incomplete tasks, initially it's assumed that all the tasks are complete
    chrm->obj[1] = 0;		//total distance travelled by each of the robots for each tasks

    //evaluation of objective 1, task completeion status
    for (int i = 0; i < tasknum; i++) {
	temp = 0;

	for (int m = 0; m < resnum; m++) {
	    sum = 0;
	    for (int j = 0; j < robonum; j++) {
		if (chrm->chrom[i * robonum + j] == 1)

		    sum += robomat[j * resnum + m];

	    }
	    if (sum < taskmat[i * resnum + m]) {
		temp = 1;
		break;
	    }
	}
	if (temp == 0)
	    taskstat[i] = 1;
	else {
	    chrm->obj[0]++;
	    taskstat[i] = 0;
	}

    }

//evaluation of objective 2, distance
    for (int i = 0; i < tasknum; i++) {
	//taskdist[idx*tasknum+i]=0;
	if (taskstat[i] == 1)
	    for (int j = 0; j < robonum; j++)
		if (chrm->chrom[i * robonum + j] == 1)
		    chrm->obj[1] += taskdistmat[i * robonum + j];


    }


}




__device__ void check(sol * chrm, curandState * state)	//Checks if multiple tasks are allocated to a single robot, resolves the ambiguity by randoming alloting one of the [previously allocated tasks. Also randomly allocates a task to a robot ewich hasn't been allocated any task previously.
{
    int temp[tasknum];		//temporary array stores the location of the  tasks allocated to a robot
    int count_num = 0;		// temporary storage for number of tasks for a robot
    int idx = threadIdx.x;

    for (int i = 0; i < robonum; i++) {
	count_num = 0;
	//printf("enter" );

	for (int j = 0; j < tasknum; j++) {

	    if (chrm->chrom[j * robonum + i] == 1) {
		//      printf("if1");
		chrm->chrom[j * robonum + i] = 0;
		count_num++;
		temp[count_num - 1] = j * robonum + i;
	    }
	}			//printf("exit check");
	if (count_num != 0)
	    chrm->chrom[temp[((int) (count_num * curand_uniform(&state[idx])))]] = 1;
	else
	    chrm->chrom[(((int) (tasknum * curand_uniform(&state[idx])))) * robonum + i] = 1;
    }

}

__device__ void mutate(sol * c, curandState * state)
{
    int mutate_ind;
    int robot_mutate = 0;
    int idx = threadIdx.x;
    for (int i = 0; i < ceil(PM * CHROMSIZE); i++) {
	mutate_ind = ((int) (CHROMSIZE * curand_uniform(&state[idx])));
	robot_mutate = mutate_ind % robonum;
	if (c->chrom[mutate_ind] == 0) {
	    for (int j = 0; j < tasknum; j++)
		c->chrom[j * robonum + robot_mutate] = 0;
	    c->chrom[mutate_ind] = 1;
	} else {
	    c->chrom[mutate_ind] = 0;
	    c->chrom[((int) (tasknum * (curand_uniform(&state[idx])))) * robonum + robot_mutate] = 1;
	}
    }


}

__device__ int compare_min(double *first, double *second, int n)
{
    /*compares two n-dimensional vectors of objective values for minimization problem
       returns 1 if first dominates second
       -1 if second dominates first, and 0 otherwise */
    int obj = 0;
    int deflt = 0;
    int current;

    do {
	if (*first < *second)
	    current = 1;
	else if (*second < *first)
	    current = -1;
	else
	    current = 0;
	if ((current) && (current == -deflt)) {
	    return 0;
	}
	if (current != 0) {
	    deflt = current;
	}
	obj++;
	*first++;
	*second++;
    } while (obj < n);


    return deflt;
}

__device__ int equal(double *first, double *second, int n)
{
    /* checks to n-dimensional vectors of objectives to see if they are identical
       returns 1 if they are, 0 otherwise */
    int obj = 0;
    do {
	if (*first != *second)
	    return 0;
	*first++;
	*second++;
	obj++;
    } while (obj < n);
    return 1;
}


__device__ void arc1hive_soln(sol * s, int *grid_pop)
{
    /* given a solution s, add it arc1hive if
       a. the arc1hive is empty
       b. the arc1hive is not full and s is not dominated or equal to anything currently in the arc1hive
       c. s dominates anything in the arc1hive 
       d. the arc1hive is full but s is non dominated and is in a no more crowded square than atleast one solution
       in additin, maintain the arc1hive such that all solutions are non-ddminated */

    int i;
    int repl;
    int yes = 0;
    int most;
    int result;
    int join = 0;
    int old_arc1length;
    int set = 0;

    int tag[MAX_ARC];
    sol *tmp;
    if (!(tmp = (sol *) malloc(MAX_ARC * sizeof(sol)))) {
	printf("Out of memory for thread %d for size of %lf\n", threadIdx.x, MAX_ARC * sizeof(sol));
	return;
    }

    for (i = 0; i < MAX_ARC; i++) {
	tag[i] = 0;
    }

    if (arc1length == 0) {
	add_to_arc1hive(s);
	return;
    }
    i = 0;
    result = 0;
    while ((i < arc1length) && (result != -1)) {
	result = equal(s->obj, (&arc1[i])->obj, OBJNUM);
	if (result == 1)
	    break;
	//MINIMIZE MAXIMIZE
	if (MINMAX == 0)
	    result = compare_min(s->obj, (&arc1[i])->obj, OBJNUM);
	/* else
	   result =compare_max(s->obj,(&arc1[i]->obj,OBJNUM); */
	if ((result == 1) && (join == 0)) {
	    count = 0;
	    arc1[i] = *s;
	    join = 1;
	} else if (result == 1) {
	    tag[i] = 1;
	    set = 1;
	}

	i++;
    }

    old_arc1length = arc1length;
    if (set == 1) {
	for (i = 0; i < arc1length; i++) {
	    tmp[i] = arc1[i];
	}
	arc1length = 0;

	for (i = 0; i < old_arc1length; i++) {
	  if (tag[i] != 1) {
		count = 0;
		arc1[arc1length] = tmp[i];
		arc1length = arc1length + 1;
	    }
	}
    }
    if ((join == 0) && (result == 0))	//i.e. solution is non dominated by the list
    {
	if (arc1length == MAX_ARC) {
	    most = grid_pop[s->grid_loc];
	    for (i = 0; i < arc1length; i++) {
		if ((grid_pop[(&arc1[i])->grid_loc]) > most);
		{
		    most = grid_pop[(&arc1[i])->grid_loc];
		    repl = i;
		    yes = 1;
		}
	    }
	    if (yes) {
		count = 0;
		arc1[repl] = *s;
	    }
	} else {
	    add_to_arc1hive(s);
	}
    }
    free(tmp);
}

__device__ int find_loc(double *eval, double *gl_offset, double *gl_range)
{
    //finds the grid location of a solution given a vector of its objective values

    int loc = 0;
    int d;
    int n = 1;
    int i;

    int inc[OBJNUM];
    double width[OBJNUM];

    //if solution is out of range of any objective, return 1 more than the maximum possible grid location number
    for (i = 0; i < OBJNUM; i++) {
	if ((eval[i] < gl_offset[i]) || (eval[i] > gl_offset[i] + gl_range[i]))
	    return ((int) pow((double) 2, (OBJNUM * DEPTH)));
    }

    for (i = 0; i < OBJNUM; i++) {
	inc[i] = n;
	n *= 2;
	width[i] = gl_range[i];
    }
    for (d = 1; d <= DEPTH; d++) {
	for (i = 0; i < OBJNUM; i++) {
	    if (eval[i] < width[i] / 2 + gl_offset[i])
		loc += inc[i];
	    else
		gl_offset[i] += width[i] / 2;
	}
	for (i = 0; i < OBJNUM; i++) {
	    inc[i] *= (OBJNUM * 2);
	    width[i] /= 2;
	}
    }
    return loc;
}

__device__ void update_grid(sol * s, double *gl_offset, double *gl_range, double *gl_largest, int *grid_pop)
{

    int a, b;
    int square;
    double offset[OBJNUM];
    double largest[OBJNUM];
    double sse;
    double product;

    for (a = 0; a < OBJNUM; a++) {
	offset[a] = LARGE;
	largest[a] = -LARGE;
    }
    for (b = 0; b < OBJNUM; b++) {
	for (a = 0; a < OBJNUM; a++) {
	    if ((&arc1[a])->obj[b] < offset[b])
		offset[b] = (&arc1[a])->obj[b];
	    if ((&arc1[a])->obj[b] > largest[b])
		largest[b] = (&arc1[a])->obj[b];
	}
    }

    for (b = 0; b < OBJNUM; b++) {
	if (s->obj[b] < offset[b])
	    offset[b] = s->obj[b];
	if (s->obj[b] > largest[b])
	    largest[b] = s->obj[b];
    }

    sse = 0;
    product = 1;

    for (a = 0; a < OBJNUM; a++) {
	sse += ((gl_offset[a] - offset[a]) * (gl_offset[a] = offset[a]));
	sse += ((gl_largest[a] - largest[a]) * (gl_largest[a] - largest[a]));
	product *= gl_range[a];
    }

    if (sse > (0.1 * product * product)) {
	change[threadIdx.x]++;

	for (a = 0; a < OBJNUM; a++) {
	    gl_largest[a] = largest[a] + 0.2 * largest[a];
	    gl_offset[a] = offset[a] + 0.2 * offset[a];
	    gl_range[a] = gl_largest[a] - gl_offset[a];
	}

	for (a = 0; a < (int) pow((double) 2, (OBJNUM * DEPTH)); a++) {
	    grid_pop[a] = 0;
	}

	for (a = 0; a < arc1length; a++) {
	    square = find_loc((&arc1[a])->obj, gl_offset, gl_range);
	    (&arc1[a])->grid_loc = square;
	    grid_pop[square]++;
	}
    }
    square = find_loc(s->obj, gl_offset, gl_range);
    s->grid_loc = square;
    grid_pop[(int) pow((double) 2, (OBJNUM * DEPTH))] = -5;
    grid_pop[square]++;
}


__device__ int compare_to_arc1hive(sol * s)
{
    int i = 0;
    int result = 0;

    while ((i < arc1length) && (result != 1) && (result != -1)) {
	if (MINMAX == 0)
	    result = compare_min(s->obj, (&arc1[i])->obj, OBJNUM);
/*		else
			result=compare_max(s->obj,(&arc1[i]->obj,OBJNUM); */
	i++;
    }
    return result;
}




__host__ void print_sol(sol * s, int *arclen, int chrom = 1, int obj = 0, int loc = 0, int thread = 0)
{				//prints the entire genetic material, correponsing objectives & grid loc for a given ptr

    int num = arclen[thread];	// length of the archive for that particular thread

    for (int i = 0; i < num; i++) {

	fprintf(file, "element no. %d\n", i + 1);
	if (chrom == 1) {
	    fprintf(file, "chromosome \n");
	    for (int k = 0; k < tasknum; k++) {
		for (int j = 0; j < robonum; j++)
		    fprintf(file, "%d ", s[thread * MAX_ARC + i].chrom[k * robonum + j]);
		fprintf(file, "\n");
	    }
	}

	if (obj == 1) {
	    fprintf(file, "objective values \n");
	    fprintf(file, "obj 1: %lf\nobj 2: %lf\n", s[thread * MAX_ARC + i].obj[0], s[thread * MAX_ARC + i].obj[1]);
	}

	if (loc == 1) {
	    fprintf(file, "grid location: %d\n", s[thread * MAX_ARC + i].grid_loc);

	}

    }

}





__global__ void in(int *d_in1, int *d_in2, int *d_in3,sol * d_archive, curandState * state, int *d_arclen, int *d_iter,double *initial_fitness)
{
    sol c[MAX_POP];
    sol m[MAX_POP];
    int i;
    int idx = threadIdx.x;
    int result;
    double gl_offset[OBJNUM];
    double gl_range[OBJNUM];
    double gl_largest[OBJNUM];
    int grid_pop[MAX_LOC];

    curand_init(idx, idx, idx, &state[idx]);
    init(c, state);
    check(c, state);
    evaluate(c, d_in1, d_in2, d_in3);
    for(int i=0;i<OBJNUM;i++)
	initial_fitness[idx*OBJNUM+i]=c->obj[i];
	
    add_to_arc1hive(c);

    i = 0;
    iterdone[threadIdx.x] = i;
    while (i++ < ITERATIONS && count++ < CONVERGENCE) {
	iterdone[threadIdx.x] = i;
/*	if (i % 100 == 0)
	    printf("\n%d -> %d", idx, i);
*/	*m = *c;
	mutate(m, state);
	evaluate(m, d_in1, d_in2, d_in3);
	if (MINMAX == 0)
	    result = compare_min(c->obj, m->obj, OBJNUM);

	if (result != 1)	//if mutant is not dominated by current (else discard it )
	{
	    if (result == -1)	//if mutant dominates current
	    {
		update_grid(m, gl_offset, gl_range, gl_largest, grid_pop);	//calculate the grid location of mutant solution and renormalize arc1hive if necessary
		arc1hive_soln(m, grid_pop);	//update the arc1hive by removing all dominated individuals
		*c = *m;	//replace c with m
	    }

	    else if (result == 0) {
		result = compare_to_arc1hive(m);
		if (result != -1) {
		    update_grid(m, gl_offset, gl_range, gl_largest, grid_pop);
		    arc1hive_soln(m, grid_pop);
		    if ((grid_pop[m->grid_loc] <= grid_pop[c->grid_loc]) || (result == 1)) {
			*c = *m;
		    }
		  
		}
	    }
	}

    	__syncthreads();
    }

    if (idx == 0) {
	d_arclen[idx] = arc1length;
	for (int i = 0; i < arc1length; i++) {
	    d_archive[i] = arc1[i];
	}
    }
    d_iter[idx] = iterdone[idx];	//for printing the no. of iterations undertaken by each thread
}

int main()
{
    clock_t start = clock();
    int *d_in1;
    int *d_in2;
    int *d_in3;
    int *d_iter;
    sol *d_archive;
    int *d_arclen;		//length of archives of each of threads
    double *d_initial_fitness;
    sol h_archive[MAX_ARC];
    int h_iterdone[THRNUM];
    double h_initial_fitness[OBJNUM*THRNUM];
    int h_arclen[THRNUM];	//lengths of the archives for each of the elements (host)
    int ARRAY_BYTES1 = robonum * resnum * sizeof(int);
    int ARRAY_BYTES2 = tasknum * resnum * sizeof(int);
    int ARRAY_BYTES3 = robonum * tasknum * sizeof(int);
    int ARRAY_BYTES4 = THRNUM * sizeof(int);
    int ARCSIZE = sizeof(sol) * MAX_ARC;
    int max_iter=0;
    char file_w_name[100];
    char file_res_robo[100];
    char file_res_task[100];
    char file_task_dist[100];
    curandState *d_state;
    sprintf(file_res_robo,"Resource_Robot_%d_%d_1.txt",robonum,tasknum);
    sprintf(file_res_task,"Resource_Task_%d_%d_1.txt",robonum,tasknum);
    sprintf(file_task_dist,"Task_Robot_Distance_%d_%d_1.txt",robonum,tasknum);
    FRM = fopen(file_res_robo, "r");

    input(FRM, robonum * resnum, RM);
    fclose(FRM);

    FTM = fopen(file_res_task, "r");
    input(FTM, tasknum * resnum, TM);
    fclose(FTM);

    FTDM = fopen(file_task_dist, "r");
    input(FTDM, robonum * tasknum, TDM);
    fclose(FTDM);

    cudaMalloc((void **) &d_state, THRNUM * sizeof(curandState));
    cudaMalloc((void **) &d_in1, ARRAY_BYTES1);
    cudaMalloc((void **) &d_in2, ARRAY_BYTES2);
    cudaMalloc((void **) &d_in3, ARRAY_BYTES3);
    cudaMalloc((void **) &d_arclen, ARRAY_BYTES4);
    cudaMalloc((void **) &d_iter, THRNUM * sizeof(int));
    cudaMalloc((void **) &d_archive, ARCSIZE);
    cudaMalloc((void **) &d_initial_fitness,sizeof(double)*OBJNUM*THRNUM);
    cudaMemcpy(d_in1, RM, ARRAY_BYTES1, cudaMemcpyHostToDevice);
    cudaMemcpy(d_in2, TM, ARRAY_BYTES2, cudaMemcpyHostToDevice);
    cudaMemcpy(d_in3, TDM, ARRAY_BYTES3, cudaMemcpyHostToDevice);

    in <<< dim3(1, 1, 1), dim3(THRNUM, 1, 1) >>> (d_in1, d_in2, d_in3, d_archive, d_state, d_arclen, d_iter,d_initial_fitness);
    clock_t end = clock();
    float sec = (float) (end - start) / CLOCKS_PER_SEC;

    cudaMemcpy(h_archive, d_archive, ARCSIZE, cudaMemcpyDeviceToHost);
    cudaMemcpy(h_arclen, d_arclen, ARRAY_BYTES4, cudaMemcpyDeviceToHost);
    cudaMemcpy(h_iterdone, d_iter, sizeof(int) * THRNUM, cudaMemcpyDeviceToHost);
    cudaMemcpy(h_initial_fitness,d_initial_fitness,sizeof(double)*OBJNUM*THRNUM,cudaMemcpyDeviceToHost);
    for (int i=0;i<THRNUM;i++)
	if(max_iter<h_iterdone[i])
		max_iter=h_iterdone[i];
		

    sprintf(file_w_name, "paes_parallel_%d_%d_%d_%d_%d_%d_%f_%d.txt",robonum, tasknum, resnum, max_iter,THRNUM,ITERATIONS,PM,CONVERGENCE);
    file = fopen(file_w_name, "w");
    fprintf(file,"Initial fitness values for each of the threads is as follows\n");
    for(int i=0;i<THRNUM;i++)
	fprintf(file,"%f , %f\n",h_initial_fitness[i*OBJNUM+0],h_initial_fitness[i*OBJNUM+1]);
    fprintf(file,"\n");
    fprintf(file, "Archive is now\n");
    for(int i=0;i<h_arclen[0];i++)
		fprintf(file,"%f , %f\n",h_archive[i].obj[0],h_archive[i].obj[1]);
//  print_sol(h_archive, h_arclen, 0, 1, 1, 0);
//   fprintf(file, "\n");
    fprintf(file,"\n\n");
    fprintf(file,"Maximimum no. of iterations undertaken by a thread : %d\n",max_iter);
    fprintf(file,"Convergence Critera : %d\n",CONVERGENCE);
    fprintf(file,"Time Taken : %f\n",sec);


//   printf("\n Number of iterations undertaken by each thread is as follows for a Convergence of %dn", CONVERGENCE);
/*    for (int i = 0; i < THRNUM; i++)
	printf("Thread no. %d \t No. of iteration : %d\n", i, h_iterdone[i]);
*/
    fclose(file);
    cudaFree(d_in1);
    cudaFree(d_in2);
    cudaFree(d_in3);
    cudaFree(d_state);
    cudaFree(d_archive);


//  printf("\nThe time elapsed %f\n", sec);
    return 0;
}

